/*************************************************************************
* Copyright (C) 2013 Intel Corporation
*
* Licensed under the Apache License,  Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* 	http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law  or agreed  to  in  writing,  software
* distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the  specific  language  governing  permissions  and
* limitations under the License.
*************************************************************************/

/*
//
//  Purpose:
//     Cryptography Primitive.
//     Digesting message according to SM3
//
//  Contents:
//     cpFinalizeSM3()
//
//
*/

#include "owndefs.h"
#include "owncp.h"
#include "hash/pcphash.h"
#include "hash/pcphash_rmf.h"
#include "pcptool.h"
#include "hash/sm3/pcpsm3stuff.h"

/* clang-format off */
IPP_OWN_DEFN(void, cpFinalizeSM3, (DigestSHA1 pHash,
                                   const Ipp8u* inpBuffer,
                                   int inpLen,
                                   Ipp64u processedMsgLen))
/* clang-format on */
{
    /* local buffer and it length */
    Ipp8u buffer[MBS_SM3 * 2];
    int bufferLen = inpLen < (MBS_SM3 - (int)MLR_SM3) ? MBS_SM3 : MBS_SM3 * 2;

    /* copy rest of message into internal buffer */
    CopyBlock(inpBuffer, buffer, inpLen);

    /* pad message */
    buffer[inpLen++] = 0x80;
    PadBlock(0, buffer + inpLen, (cpSize)(bufferLen - inpLen - (int)MLR_SM3));

    /* put processed message length in bits */
    processedMsgLen                     = ENDIANNESS64(processedMsgLen << 3);
    ((Ipp64u*)(buffer + bufferLen))[-1] = processedMsgLen;

    /* complete hash computation */
    UpdateSM3(pHash, buffer, bufferLen, sm3_cnt);
}
