/*************************************************************************
* Copyright (C) 2002 Intel Corporation
*
* Licensed under the Apache License,  Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* 	http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law  or agreed  to  in  writing,  software
* distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the  specific  language  governing  permissions  and
* limitations under the License.
*************************************************************************/

/*
//               Intel(R) Cryptography Primitives Library
//
//   Purpose: Describes the version of ippCP
//
*/


#include "ippver.h"
#define BUILD()            1043
#define VERSION()          BASE_VERSION(), BUILD()
#define STR_FILE_VERSION() STR_BASE_VERSION() "," STR(BUILD())

/* ////////////////////////// End of file "pcpver.h" ///////////////////////// */
